/*
 * A class that supplies a sequence of bits.
 * 2021-03-27: Created by Abdalla Ahmed from an earlier version.
 *
 */

#ifndef BITSTREAM_H
#define BITSTREAM_H

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <random>

std::random_device rd;                                                          // Will be used to obtain a seed for the random number engine
std::mt19937 rnd(rd());

inline uint32_t bitReverse(uint32_t n) {
    n = (n << 16) | (n >> 16);
    n = ((n & 0x00ff00ff) << 8) | ((n & 0xff00ff00) >> 8);
    n = ((n & 0x0f0f0f0f) << 4) | ((n & 0xf0f0f0f0) >> 4);
    n = ((n & 0x33333333) << 2) | ((n & 0xcccccccc) >> 2);
    n = ((n & 0x55555555) << 1) | ((n & 0xaaaaaaaa) >> 1);
    return n;
}

inline uint32_t vdc2sobol(uint32_t x) {
    uint32_t y(x);
    y ^= (y & 0x0000FFFF) << 16;
    y ^= (y & 0x00FFFF00) <<  8;
    y ^= (y & 0x0F0F0F0F) <<  4;
    y ^= (y & 0x33333333) <<  2;
    y ^= (y & 0x55555555) <<  1;
    return y;
}

/******************************************************************************/
/******************************************************************************/

class BitStream {
public:
    virtual int getBit() { return 0; };
};

/******************************************************************************/
/******************************************************************************/
class BitStreamFixed : public BitStream {
private:
    int bit;
public:
    BitStreamFixed(int fixedBit) { bit = fixedBit & 1; };
    int getBit() { return bit; };
};

/******************************************************************************/
/******************************************************************************/
class BitStreamBitMap : public BitStream {
private:
    int bitmap[32 * 32];
    int seqNo;
public:
    BitStreamBitMap(char *fileName);
    int getBit();
};

BitStreamBitMap::BitStreamBitMap(char *fileName) {
    seqNo = 0;
    FILE *file = fopen(fileName, "r");
    if (!file) {
        fprintf(stderr, "failed to open image file %s\n", fileName);
        exit(1);
    }
    int w, h;
    fscanf(file, "P1 %d %d", &w, &h);                                           // Assuming a PBM file
    if (w != 32 || h != 32) {
        fprintf(stderr, "Image size must be 32 (width) x 32 (height)\n");
        exit(2);
    }
    for (int i = 0; i < 32 * 32; i++) {
        char c;
        fscanf(file, " %c", &c);
        bitmap[i] = c == '1' ? 1 : 0;
        if (feof(file)) {
            fprintf(stderr, "Failed to read all data from %s\n", fileName);
            exit(3);
        }
    }
    fprintf(stderr, "Input image:\n");
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            fprintf(stderr, " %c", bitmap[y * 32 + x] ? '#' : '.');
        }
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "\nScrambled image:\n");
    for (int y = 0; y < 32; y++) {
        for (int x = 0; x < 32; x++) {
            fprintf(stderr, " %c", getBit() ? '#' : '.');
        }
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "seqNo = %d\n", seqNo);
    seqNo = 0;
    fclose(file);
}

/******************************************************************************/

int BitStreamBitMap::getBit() {
    uint32_t x, y;
    x = bitReverse(seqNo++);
    y = vdc2sobol(x);
    x >>= (32 - 5);                                                          // Bring it to the range (0, 31)
    y >>= (32 - 5);
    return bitmap[y * 32 + x];
}

/******************************************************************************/
/******************************************************************************/
class BitStreamRandom : public BitStream {
public:
    int getBit() { return rnd() & 1; }
};

/******************************************************************************/
/******************************************************************************/
class BitStreamList : public BitStream {                                        // Delivers a stored list of bits
private:
    int bitCount;                                                               // Number of stored bits.
    int seqNo;                                                                  // Sequence number of retrieved bits.
    std::vector<uint32_t> bits;                                                 // List of words containing the bits.
public:
    BitStreamList() : bitCount(0), seqNo(0) { bits.resize(64); };               // This should be more than enough.
    void push_back(int bit);                                                    // Insert a bit
    int getBit();
};

/******************************************************************************/
void BitStreamList::push_back(int bit) {
    int wordIndex = bitCount >> 5;                                              // Which word will accept the received bit?
    int bitIndex = bitCount & 32;
    uint32_t mask = (1 << bitIndex) - 1;
    if (wordIndex > bits.size()) {
        bits.resize(2 * bits.size());
    }
    bits[wordIndex] = (bits[wordIndex] & mask) | ((bit & 1) << bitIndex);
    bitCount++;
}

/******************************************************************************/
int BitStreamList::getBit() {
    int bit(0);
    if (seqNo < bitCount) {
        bit = (bits[seqNo >> 5] >> (seqNo & 32)) & 1;
    }
    seqNo++;
    return bit;
}

/******************************************************************************/
/******************************************************************************/
class BitStream32 : public BitStream {                                          // Cycle through bits of a 32-bit int. Useful for enumeration.
private:
    uint32_t bits, seqNo;
public:
    BitStream32(uint32_t n) : bits(n), seqNo(0) { };
    int getBit() { return (bits >> ((seqNo++) & 31)) & 1; };
    BitStream& reset(uint32_t n) {
        bits = n; seqNo = 0;
        //fprintf(stderr, "BitStream reset to %u\n", n);
        return *this;
    }
};


#endif
