/*
 * Binary matrix manipulation.
 * The matrix is primarily encoded by unsigned integers representing rows.
 * The least-significant bit stands for the left-most column.
 * 2022-05-26: Created by Abdalla Ahmed.
 * 2022-06-08: Moved classes to this header file.
 *
 */

#include "bitstream.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <vector>
#include <string>

// =============================================================================
// Data types
// =============================================================================

typedef std::string String;

struct Point {
    uint32_t x, y;
};

// =============================================================================
// Global parameters
// =============================================================================

bool verbose = true;
const char code[2] = {'.', '1'};


// =============================================================================
// Utilities
// =============================================================================

inline String binaryString(uint32_t v, int bits) {
    String result;
    result.resize(bits);
    for (int i = bits - 1; i >= 0; i--) {
        result[i] = code[(v >> i) & 1];
    }
    return result;
}

inline uint32_t lsb(uint32_t x) {                                               // Isolate least-significant bit
    return x - ((x - 1) & x);
}

// =============================================================================
// Matrix Class
// =============================================================================

class Matrix {
private:
    int m;                                                                      // Size
    uint32_t rows[32];
    void printInversionStep(uint32_t *in, uint32_t *out, FILE *file);
public:
    enum MODE {GENERAL, PROGRESSIVE, LTRIANGULAR};
    Matrix() { m = 0; };                                                        // Create an empty matrix.
    Matrix(int m);                                                              // Create a random regular matrix of size m.
    Matrix(int m, uint32_t bits);                                               // Populate bits from supplied int. Useful for enumeration.
    Matrix(int m, BitStream *bs, MODE mode = PROGRESSIVE);                      // Create a matrix by reading the optional bits from the bitstream.
    Matrix(FILE *file);                                                         // Reads the matrix from a file. Format is m followed by row-major list of entries. Diagonal entries are adjusted as needed.
    Matrix(int m, const uint32_t *inputRows);                                   // Copy the matrix from a given list of unsigned integers.
    Matrix(Matrix A, Matrix B, int rowsFromA);                                  // Create a matrix from the first rowsFromA rows from A, and the subsequent rows from B, assuming the size of A.
    Matrix inverse(FILE *log = NULL);                                            // Return the inverse matrix.
    Matrix dyadicPair(BitStream *bs);                                           // Create a matrix that makes a dyadic net when paired with this.
    Matrix progressivePair(BitStream *bs);                                      // Creates a matrix that can be paired with this one to make a (0, 2)-sequence, taking the bits bits from bs.
    void extend(int newSize, BitStream *bs);                                    // Add rows and columns from bs to a regular matrix and adjust the parity to make the matrix progressive.
    void coExtend(Matrix &pair, int newSize, BitStream *bs);                    // Extend both matrices while maintaining their co-progression.
    void print(FILE *file) const;                                               // Prints the matrix to a file.
    void printC(FILE *file = stdout) const;                                     // Print as a C array of columns.
    void regularize();                                                          // Make the matrix progressively regular by computing the diagonal bits.
    bool isSingular();                                                          // Return true if the matrix is singular.
    int rank();                                                                 // Rank of matrix.
    int rankLead();                                                             // Rank of leading rows, no reordering.
    bool luDecompose(Matrix &L, Matrix &U, FILE *log = NULL);                   // Compute LU Decomposition, or return false and empty matrices on failure.
    bool lvDecompose(Matrix &L, Matrix &V);                                     // Similar to LU, but using any regular matrix, as if the upper factor is column-shuffled.
    Matrix reduce();                                                            // Factor out the lower matrix and return the v part.
    int getm() const { return m; }
    uint32_t getRow(int i) { return rows[i]; };
    inline int getBit(int row, int col) const {
        return (rows[row] >> col) & 1;
    };
    String getString();
    bool isDyadicPair(Matrix other);                                            // Checks whether matrix other can make a (0, m, 2)-net with this.
    bool isProgressivePair(Matrix other);                                       // Checks whether matrix other can make a (0, 2)-sequence with this.
    int crossRank(Matrix other);                                                // Return smallest rank of hybrid between this and other.
    Matrix operator+(const Matrix &other);
    Matrix operator*(const Matrix &other);                                      // Multiply with a matrix on the right.
    Matrix operator*=(const Matrix &other) {
        *this = *this * other; return *this;
    }
    uint32_t toUint32_t();                                                      // For m < 6
};

// -----------------------------------------------------------------------------


Matrix::Matrix(FILE *file) {
    for (int i = 0; i < 32; i++) {
        rows[i] = 0;
    }
    fscanf(file, " %d", &m);
    for (int row = 0; row < m; row++) {
        for (int col = 0; col < m; col++) {
            char c;
            fscanf(file, " %c", &c);
            if (feof(file)) {
                fprintf(stderr, "Failed to read matrix\n");
                exit(1);
            }
            int x = (c == '1') ? 1 : 0;
            rows[row] |= x << col;
        }
    }
}

// -----------------------------------------------------------------------------

Matrix::Matrix(int m) : m(m) {
    uint32_t mask = (1ull << m) - 1;
    uint32_t reduced[m];
    while (!rows[0]) { rows[0] = rnd() & mask; }
    reduced[0] = rows[0];
    for (int row = 1; row < m; row++) {
        reduced[row] = 0;
        while (!reduced[row]) {
            reduced[row] = rows[row] = rnd() & mask;                            // Pick a random row
            for (int i = 0; i < row; i++) {
                if (reduced[row] & lsb(reduced[i])) {                           // If it contain the bit of the i'th row
                    reduced[row] ^= reduced[i];                                 // Then remove that component
                }
            }
        }
    }
}
// -----------------------------------------------------------------------------

Matrix::Matrix(int m, const uint32_t *inputRows) : m(m) {                       // We make no error checking
    uint32_t mask = (1ull << m) - 1;
    for (int i = 0; i < m; i++) {
        rows[i] = inputRows[i] & mask;
    }
}

// -----------------------------------------------------------------------------

Matrix::Matrix(int m, uint32_t bits) : m(m) {                                   // We make no error checking
    uint32_t mask = (1ull << m) - 1;
    for (int i = 0; i < m; i++) {
        rows[i] = (bits >> (i * m)) & mask;
    }
}

// -----------------------------------------------------------------------------
// How bits are indexed; dots refer to evaluated bits:
//
//  GENERAL      PROGRESSIVE         LTRIANGULAR
//
//  abcd            .bej                ....
//  efgh            a.fk                a...
//  ijkl            cd.l                bc..
//  mnop            ghi.                def.
// -----------------------------------------------------------------------------

Matrix::Matrix(int size, BitStream *bs, MODE mode) {
    m = size;
    for (int row = 0; row < 32; row++) rows[row] = 0;                           // Clear rows to simplify bit insertion.
    switch (mode) {
        case GENERAL: {
            for (int i = 0; i < m; i++) {
                for (int j = 0; j < m; j++) {
                    rows[i] |= bs->getBit() << j;
                }
            }
            break;
        }
        case LTRIANGULAR: {
            for (int i = 0; i < m; i++) {
                rows[i] |= 1 << i;
                for (int j = 0; j < i; j++) {
                    rows[i] |= bs->getBit() << j;
                }
            }
            break;
        }
        case PROGRESSIVE:
        default: {
            for (int stripe = 1; stripe < m; stripe++) {
                for (int col = 0; col < stripe; col++) {
                    rows[stripe] |= bs->getBit() << col;
                }
                for (int row = 0; row < stripe; row++) {
                    rows[row] |= bs->getBit() << stripe;
                }
            }
            regularize();
        }
    }
}

// -----------------------------------------------------------------------------

Matrix::Matrix(
    Matrix A, Matrix B, int rowsFromA
) {
    m = A.m;
    for (int i = 0; i < rowsFromA; i++) {
        rows[i] = A.rows[i];
    }
    for (int i = 0; i < m - rowsFromA; i++) {
        rows[i + rowsFromA] = B.rows[i];
    }
}

// -----------------------------------------------------------------------------

void Matrix::regularize() {
    uint32_t tmp[m];                                                            // A temporary copy for processing
    for (int i = 0; i < m; i++) {
        tmp[i] = rows[i];
    }
    for (int i = 0; i < m; i++) {                                               // Iterate through the rows.
        if (((tmp[i] >> i) & 1) == 0) {                                         // Is the diagonal bit unset?
            tmp[i]  |= 1 << i;                                                  // Set it in the processed matrix,
            rows[i] ^= 1 << i;                                                  //  and also toggle the control bit in the original matrix.
        }
        for (int k = i + 1; k < m; k++) {                                       // Iterate through subsequent rows
            if ((tmp[k] >> i) & 1) {                                            // Is a bit set below the current diagonal bit?
                tmp[k] ^= tmp[i];                                               // Clear it by adding the reference row
            }
        }
    }
}

// -----------------------------------------------------------------------------

Matrix Matrix::dyadicPair(BitStream *bs) {
    Matrix result;
    result.m = m;
    uint32_t mask = (1ull << m) - 1;
    for (int row = 0; row < m; row++) {
        // =====================================================================
        // Build a tmp matrix to apply Gaussian elimination
        // =====================================================================
        uint32_t reduced[m];
        for (int i = 0; i < m - 1 - row; i++) {
            reduced[i] = rows[i];                                               // Populate leading rows from this
        }
        for (int i = 0; i < row; i++) {
            reduced[m - 1 - row + i] = result.rows[i];                          // Populate already computed rows
        }
        // =====================================================================
        // Apply Gaussian elimination, and take note of column ownership
        // =====================================================================
        uint32_t occupied = lsb(reduced[0]);                                    // Bit mask to indicate which bits are taken by leading rows.
        for (int i = 1; i < m - 1; i++) {                                       // Iterate through rows of reduced matrix
            for (int k = 0; k < i; k++) {                                       // Iterate through preceding rows
                if (reduced[i] & lsb(reduced[k])) {                             // If it contain the bit of the k'th row
                    reduced[i] ^= reduced[k];                                   // Then remove that component
                }
            }
            occupied |= lsb(reduced[i]);
        }
        uint32_t free = ~occupied & mask;                                       // Identify free bit
        // =====================================================================
        // Compute last independent row
        // =====================================================================
        result.rows[row] = 0;                                                   // Initialize to zeros for reading bs.
        for (int j = 0; j < m; j++) {                                           // Iterate through columns (bits).
            if (!((free >> j) & 1)) {                                           // Skip the free bit.
                result.rows[row] |= bs->getBit() << j;                          // Insert a bit from stream.
            }
        }
        reduced[m - 1] = result.rows[row];                                      // Insert in reduced matrix.
        for (int k = 0; k < m - 1; k++) {                                       // Iterate through preceding rows.
            if (reduced[m - 1] & lsb(reduced[k])) {                             // If it contain the bit of the k'th row.
                reduced[m - 1] ^= reduced[k];                                   // Then remove that component.
            }
        }
        if (!reduced[m - 1]) {                                                  // If matrix is singular,
            result.rows[row] ^= free;                                           // Then toggle the associated bit.
        }
    }
    return result;
}

// -----------------------------------------------------------------------------

Matrix Matrix::progressivePair(BitStream *bs) {
    Matrix result(m, bs, LTRIANGULAR);                                          // Populate lower triangle from bitstream
    for (int size = 2; size <= m; size++) {                                     // Iterate through progressive sizes
        result.m = size;                                                        // Restrict to current size
        for (int row = 0; row < size; row++) {                                  // Iterate through rows of the restricted matrix
            Matrix hybrid(result, *this, row + 1);                              // First (row + 1) rows of result and rest from this
            if (hybrid.isSingular()) {                                          // Is it singular?
                result.rows[row] ^= 1 << (size - 1);                            // Then flip the bit to make it regular.
            }                                                                   // Note that this is inclusive of all rows from new matrix
        }
    }
    return result;
}

// -----------------------------------------------------------------------------

void Matrix::extend(int newSize, BitStream *bs) {
    if (isSingular() || newSize > 32) return;
    for (int stripe = m; stripe < newSize; stripe++) {
        rows[stripe] = 0;
        for (int col = 0; col < stripe; col++) {
            rows[stripe] |= bs->getBit() << col;
        }
        for (int row = 0; row < stripe; row++) {
            rows[row] |= bs->getBit() << stripe;
        }
        m++;
        if (isSingular()) {
            rows[stripe] ^= 1 << stripe;
        }
    }
}

// -----------------------------------------------------------------------------

void Matrix::coExtend(Matrix &pair, int newSize, BitStream *bs) {
    if (isSingular() || !isDyadicPair(pair) || newSize > 32) {
        return;
    }
    for (int stripe = m; stripe < newSize; stripe++) {
        // =====================================================================
        // Extend self
        // =====================================================================
        rows[stripe] = 0;
        for (int col = 0; col < stripe; col++) {
            rows[stripe] |= bs->getBit() << col;
        }
        for (int row = 0; row < stripe; row++) {
            rows[row] |= bs->getBit() << stripe;
        }
        m++;
        if (isSingular()) {
            rows[stripe] ^= 1 << stripe;
        }
        // =====================================================================
        // Extend pair
        // =====================================================================
        pair.m = m;
        for (int row = 0; row < m - 1; row++) {                                 // Iterate through rows of pair
            Matrix hybrid(pair, *this, row + 1);                                // First (row + 1) rows of result and rest from this
            if (hybrid.isSingular()) {                                          // Is it singular?
                pair.rows[row] ^= 1 << (m - 1);                                 // Then flip the bit to make it regular.
            }
        }
        pair.rows[stripe] = 0;
        for (int col = 0; col < m - 1; col++) {
            pair.rows[m - 1] |= bs->getBit() << col;
        }
        if (pair.isSingular()) {
            pair.rows[m - 1] ^= 1 << (m - 1);
        }
    }
}

// -----------------------------------------------------------------------------

bool Matrix::isDyadicPair(Matrix B) {
    for (int rowsFromThis = 0; rowsFromThis <= m; rowsFromThis++) {
        if (Matrix(*this, B, rowsFromThis).isSingular()) {
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------------

bool Matrix::isProgressivePair(Matrix B) {
    Matrix A = *this;
    for (int size = 1; size <= m; size++) {
        A.m = B.m = size;
        if (!A.isDyadicPair(B)) {
            return false;
        }
    }
    return true;
}

// -----------------------------------------------------------------------------

void Matrix::print(FILE *file) const {
    fprintf(file, "%d\n", m);
    for (int row = 0; row < m; row++) {
        fprintf(file, "%s\n", binaryString(rows[row], m).c_str());
    }
}

// -----------------------------------------------------------------------------

void Matrix::printC(FILE *file) const {
    for (int col = 0; col < m; col++) {
        if (col % 4) { fprintf(file, "    "); }
        else         { fprintf(file, "\n    ");}
        uint32_t v(0);
        for (int row = 0; row < m; row++) {
            v |= getBit(row, col) << (31 - row);
        }
        fprintf(file, "0x%08x", v);
        if (col < m - 1) fprintf(file, ",");
    }
    fprintf(file, "\n");
}

// -----------------------------------------------------------------------------

String Matrix::getString() {
    String s;
    s.reserve(m * m);
    for (int stripe = 0; stripe < m; stripe++) {
        for (int col = 0; col < stripe; col++) {                                // Insert row from left to right
            s.push_back(code[ getBit(stripe, col) ]);
        }
        for (int row = 0; row < stripe; row++) {                                // Insert column from top to bottom
            s.push_back(code[ getBit(row, stripe) ]);
        }
        s.push_back(code[ getBit(stripe, stripe) ]);
    }
    return s;
}

// -----------------------------------------------------------------------------

inline void Matrix::printInversionStep (
    uint32_t *in, uint32_t *out, FILE *file
) {
    if (file) {
        for (int row = 0; row < m; row++) {
            fprintf(
                file, "%s|%s\n",
                binaryString( in[row], m).c_str(),
                binaryString(out[row], m).c_str()
            );
        }
        fprintf(file, "--------------------------------\n");
    }
}

// -----------------------------------------------------------------------------

Matrix Matrix::inverse(FILE *file) {
    uint32_t in[m], out[m];
    for (int i = 0; i < m; i++) {
        in[i] = rows[i];
        out[i] = 1 << i;                                                        // Identity
    }
    printInversionStep(in, out, file);
    for (int row = 0; row < m; row++) {                                         // Iterate throw rows
        uint32_t pivot = 1 << row;                                              // Diagonal bit
        if ((in[row] & pivot) == 0) {                                           // Is it missing in the current row?
            for (int i = row + 1; i < m; i++) {                                 // Scan the subsequent rows
                if (in[i] & pivot) {                                            // If a row provides this bit
                    std::swap(in[row], in[i]);                                  // Swap it with the current row; equivalent to {in[row] ^= in[i]; in[i] ^= in[row]}.
                    std::swap(out[row], out[i]);                                // Same same.
                    printInversionStep(in, out, file);                          // Report this step.
                    break;
                }
            }
        }
        if ((in[row] & pivot) == 0) {                                           // Still no luck?
            return (Matrix());                                                  // Return an empty matrix.
        }
        bool changed = false;                                                   // To track and aggregate changes.
        for (int i = row + 1; i < m; i++) {                                     // Iterate through subsequent rows
            if (in[i] & pivot) {                                                // If they contain diagonal bit ..
                in[i] ^= in[row];                                               // Eliminate it.
                out[i] ^= out[row];                                             // Same same
                changed = true;                                                 // At least one row has changed.
            }
        }
        if (changed) printInversionStep(in, out, file);
    }
    for (int row = m - 1; row >= 0; row--) {                                    // Now iterate backward to eliminate upper triangle; all diagonal bits are set!
        uint32_t pivot = 1 << row;
        bool changed = false;                                                   // To track and aggregate changes.
        for (int i = row - 1; i >= 0; i--) {
            if (in[i] & pivot) {
                in[i] ^= in[row];                                               // Now we have a 1 in diagonal entry
                out[i] ^= out[row];
                changed = true;
            }
        }
        if (changed) printInversionStep(in, out, file);
    }
    return Matrix(m, out);
}

// -----------------------------------------------------------------------------
// Check Wikipedia page https://en.wikipedia.org/wiki/LU_decomposition
// The steps include successive application of Gaussian elimination on the
// bottom half. The Upper matrix is what is left on the original matrix
// The lower half is the product of the inverse of
// the hanging-column lower-triangular matrices used for elimination.
// Many simplifications come from the binary nature of our matrices.
// Specifically:
//  1. the hanging-column matrices are inverses of themselves, as may
//     easily be seen using a single step of Gaussian elimination.
//  2. The product of these matrices is just a juxtaposition of the columns.
//     To see that, think of applying the multiplication from left to right
//     (longest to shortest column). The newly multiplied column may further
//     be decomposed as a sum of an identity matrix and the hanging column.
//     Multiplied by this hanging column uses its set bits to add the
//     corresponding columns of the multiplied matrix, and place the result
//     in this column. Now these columns contain only diagonal bits at this
//     time, hence the result is the hanging column itself added to the
//     multiplied matrix.
// -----------------------------------------------------------------------------

bool Matrix::luDecompose(Matrix &L, Matrix &U, FILE *file) {
    L.m = U.m = 0;                                                              // Assume failure
    uint32_t *in = U.rows;
    uint32_t *out = L.rows;
    for (int i = 0; i < m; i++) {
        in[i] = rows[i];                                                        // This would just undergo a half-inversion
        out[i] = 1 << i;                                                        // This would receive a modified half-inversion
    }
    printInversionStep(in, out, file);
    for (int row = 0; row < m; row++) {                                         // Iterate throw rows
        uint32_t pivot = 1 << row;                                              // Diagonal bit
        if ((in[row] & pivot) == 0) {                                           // Is it missing in the current row?
            return false;                                                       // Then the matrix is not progressive and cannot be LU-decomposed
        }
        bool changed = false;                                                   // To track and aggregate changes.
        for (int i = row + 1; i < m; i++) {                                     // Iterate through subsequent rows
            if (in[i] & pivot) {                                                // If they contain diagonal bit ..
                in[i] ^= in[row];                                               // Eliminate it.
                out[i] ^= pivot;                                                // This is the difference from Gaussian elimination: we only modify the current column.
                changed = true;                                                 // At least one row has changed.
            }
        }
        if (changed) printInversionStep(in, out, file);
    }
    L.m = U.m = m;                                                              // Now the process succeeded, validate the matrices
    return true;
}

// -----------------------------------------------------------------------------
// Very similar to LU decomposition but requires the matrix to be only regular,
// hence the upper is replaced by a column-shuffled upper.
// -----------------------------------------------------------------------------

bool Matrix::lvDecompose(Matrix &L, Matrix &V) {
    L.m = V.m = 0;                                                              // Assume failure
    uint32_t *in = V.rows;
    uint32_t *out = L.rows;
    for (int i = 0; i < m; i++) {
        in[i] = rows[i];                                                        // This would just undergo a half-inversion
        out[i] = 1 << i;                                                        // This would receive a modified half-inversion
    }
    for (int row = 0; row < m; row++) {                                         // Iterate throw rows
        if (!in[row]) {
            return false;
        }
        uint32_t pivot = lsb(in[row]);                                          // Column to be cleared in the left matrix.
        uint32_t diagonal = 1 << row;                                           // Diagonal column in the lower matrix.
        for (int i = row + 1; i < m; i++) {                                     // Iterate through subsequent rows
            if (in[i] & pivot) {                                                // If they contain diagonal bit ..
                in[i] ^= in[row];                                               // Eliminate it.
                out[i] ^= diagonal;                                             // This is the difference from Gaussian elimination: we only modify the current column.
            }
        }
    }
    L.m = V.m = m;                                                              // Now the process succeeded, validate the matrices
    return true;
}

// -----------------------------------------------------------------------------
// Just factor out the lower part.
// -----------------------------------------------------------------------------

Matrix Matrix::reduce() {
    Matrix reduced = *this;
    reduced.m = 0;
    uint32_t *in = reduced.rows;
    for (int row = 0; row < m; row++) {                                         // Iterate throw rows
        if (!in[row]) {
            return reduced;
        }
        uint32_t pivot = lsb(in[row]);                                          // Column to be cleared in the left matrix.
        for (int i = row + 1; i < m; i++) {                                     // Iterate through subsequent rows
            if (in[i] & pivot) {                                                // If they contain diagonal bit ..
                in[i] ^= in[row];                                               // Eliminate it.
            }
        }
    }
    reduced.m = m;                                                              // Now the process succeeded, validate the matrix.
    return reduced;
}


// -----------------------------------------------------------------------------

Matrix Matrix::operator+(const Matrix &other) {
    Matrix sum;
    sum.m = m;
    for (int row = 0; row < m; row++) {
        sum.rows[row] = rows[row] ^ other.rows[row];
    }
    return sum;
}

// -----------------------------------------------------------------------------

Matrix Matrix::operator*(const Matrix &other) {
    Matrix product;
    product.m = m;
    for (int row = 0; row < m; row++) { product.rows[row] = 0; }                // Initialize to zeros
    for (int row = 0; row < m; row++) {                                         // Iterate through rows of this matrix
        for (int i = 0; i < m; i++) {                                           // Iterate through entries of current row
            if ((rows[row] >> i) & 1) {
                product.rows[row] ^= other.rows[i];                             // If a bit is set, it selects the whole row of other
            }
        }
    }
    return product;
}

// -----------------------------------------------------------------------------

bool Matrix::isSingular() {
    uint32_t tmp[m];
    for (int i = 0; i < m; i++) {
        tmp[i] = rows[i];
    }
    for (int row = 0; row < m; row++) {
        uint32_t pivot = 1 << row;                                              // Diagonal bit
        if ((tmp[row] & pivot) == 0) {                                          // This is a special case that we need to handle
            for (int i = row + 1; i < m; i++) {
                if (tmp[i] & pivot) {
                    std::swap(tmp[row], tmp[i]);
                    break;
                }
            }
        }
        if ((tmp[row] & pivot) == 0) {                                          // Still no luck?
            return (true);                                                      // Then the matrix is singular
        }
        for (int i = row + 1; i < m; i++) {                                     // Now scan subsequent rows
            if (tmp[i] & pivot) {                                               // Do they have a set bit under the current diagonal bit?
                tmp[i] ^= tmp[row];                                             // Eliminate it
            }
        }
    }
    return false;                                                               // All diagonal bits defined, non-singular
}

// -----------------------------------------------------------------------------

int Matrix::rank() {
    uint32_t tmp[m];
    uint32_t mask = (1 << m) - 1;
    for (int i = 0; i < m; i++) {
        tmp[i] = rows[i] & mask;
    }
    int rank = m;
    uint32_t pivot = 1;
    for (int row = 0; row < m; row++, pivot <<= 1) {
        while (!(tmp[row] & pivot) && pivot < (1 << m)) {
            for (int i = row; i < m; i++) {                                     // Scan all rows starting from current for a pivot bit
                if (tmp[i] & pivot) {
                    std::swap(tmp[row], tmp[i]);
                    break;
                }
            }
            if (!(tmp[row] & pivot)) {                                          // Still no luck?
                rank--;                                                         // Then a rank is lost
                pivot <<= 1;                                                    // Advance to the following column
            }
        }
        if (tmp[row] & pivot) {
            for (int i = row + 1; i < m; i++) {
                if (tmp[i] & pivot) {
                    tmp[i] ^= tmp[row];
                }
            }
        }
    }
    return rank;
}

// -----------------------------------------------------------------------------

int Matrix::rankLead() {
    uint32_t tmp[m];
    uint32_t mask = (1 << m) - 1;                                               // To mask out higher bits if any
    for (int i = 0; i < m; i++) {
        tmp[i] = rows[i] & mask;
    }
    for (int row = 0; row < m; row++) {
        if (tmp[row] == 0) return row;
        uint32_t pivot = tmp[row];
        pivot -= (pivot - 1) & pivot;                                           // Remove higher set bits, if any
        for (int i = row + 1; i < m; i++) {                                     // Now scan subsequent rows
            if (tmp[i] & pivot) {                                               // Do they have a set bit under pivot column?
                tmp[i] ^= tmp[row];                                             // Eliminate it
            }
        }
    }
    return m;                                                                   // All diagonal bits defined, non-singular
}
// -----------------------------------------------------------------------------

int Matrix::crossRank(Matrix other) {
    int rank = m;
    for (int rowsFromThis = 1; rowsFromThis < m; rowsFromThis++) {
        rank = std::min(
            rank, Matrix(*this, other, rowsFromThis).rankLead()
        );
    }
    return rank;
}

// -----------------------------------------------------------------------------

uint32_t Matrix::toUint32_t() {
    uint32_t result = 0;
    uint32_t mask = (1 << m) - 1;
    for (int row = m - 1; row >= 0; row--) {
        result <<= m;
        result |= rows[row] & mask;
    }
    return result;
}

// -----------------------------------------------------------------------------
// A short-hand function for indexed matrices
// -----------------------------------------------------------------------------

inline bool isDyadicPair(int m, uint32_t x, uint32_t y) {
    return Matrix(m, x).isDyadicPair(Matrix(m, y));
};

// =============================================================================
// End of Matrix class
// =============================================================================

