/*
 * Generate xi sequence.
 *
 * Points are retrieved as 32-bit unsigned integers,
 *  interpreted as fixed-point fractions.
 *
 * 2021-03-06: Created by Abdalla Ahmed from earlier versions.
 * 2021-03-07: Rewritten as a class.
 * 2021-03-09: Retired the nest-randomized version, in favor of the elegant
 *  and neat original single-point version. Variants are found in archive.
 * 2021-03-25: Worked from the unified header.
 */

#include "xi.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <iostream>
#include <vector>
#include <wmmintrin.h>
#include <getopt.h>
#include <cmath>
#include <random>

#define Float double
std::random_device rd;                                                          // Will be used to obtain a seed for the random number engine
std::mt19937 rnd(rd());

const char *USAGE_MESSAGE = "Usage: %s [options] [a.x a.y] [p0.x p0.y]\n"
"Point coordinates, if supplied, should be 32-bit hexadecimals, "
"interpreted as fixed-point fractions\n"
"Options:\n"
"    -m <depth>           Set resolution of tests; default is 12 (4096 point)\n"
"    -d <fileName.pgm>    Print a dither matrix\n"
"    -v                   Verbosely echo progress\n"
"    -r                   Report production rate\n"
"    -R                   Report inversion rate\n"
"    -s <file.txt>        Generate samples and save them to file.txt\n"
"    -i                   Interactively test inversion\n"
"    -p                   Print generating matrices to stderr\n"
"    -c                   Print conflict radius to stderr\n"
;

int main(int argc, char** argv) {
    int opt;                                                                    // For use by getopt, the command line options utility
    srand(time(NULL));
    unsigned m = 12;
    char *ditheFileName(NULL);
    char *outputFileName(NULL);
    bool reportGenerationRate(false);
    bool reportInversionRate(false);
    bool testInversion(false);
    bool printMatrix(false);
    bool printConflictRadius(false);
    while ((opt = getopt(argc, argv, "m:d:rRs:ipc")) != -1) {
        switch (opt) {
            case 'm': m = atoi(optarg); break;
            case 'd': ditheFileName = optarg; break;
            case 'r': reportGenerationRate = true; break;
            case 'R': reportInversionRate = true; break;
            case 's': outputFileName = optarg; break;
            case 'i': testInversion = true; break;
            case 'p': printMatrix = true; break;
            case 'c': printConflictRadius = true; break;
            default: fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
        }
    }
    //Vector a = {0x80000000, 0x80000000}, p0 = {0, 0};
    Vector a = {rnd() | 0x80000000, rnd() | 0x80000000}, p0 = {rnd(), rnd()};
    if (optind <= argc - 2) {                                                    // Two arguments supplied? Interpret them as p1
        char *dummy;
        a.x = strtol(argv[optind    ], &dummy, 16);
        a.y = strtol(argv[optind + 1], &dummy, 16);
        if (optind <= argc - 4) {                                                // Two more arguments supplied? Interpret them as p0
            p0.x = strtol(argv[optind + 2], &dummy, 16);
            p0.y = strtol(argv[optind + 3], &dummy, 16);
        }
    }
    Xi xi(a, p0);
    //xi.printData(stderr);
    //Xi256 xi(a, p0);
    //Xi xiref(a, p0);
    if (printMatrix) xi.printMatrix(stderr);

    //*////////////////////////// Test Inversion ///////////////////////////////
    if (testInversion) {
        while (true) {
            uint32_t seqNo, m, retrievedSeqNo, mask;
            printf("Enter seqNo and m separated by a space: \n");
            scanf(" %u %u", &seqNo, &m);
            if (feof(stdin)) break;
            mask = (1 << m) - 1;
            Point p = xi[seqNo];
            Vector stratum = p >> (32 - m);
            retrievedSeqNo = xi.getSeqNo(stratum.x, stratum.y, m);
            printf (
                "Generated point = (%08X, %08X), Retrieved seqNo = %u\n",
                p.x, p.y, retrievedSeqNo
            );
        }
    }
    //////////////////////////////////////////////////////////////////////////*/

    //*////////////// Print a dither matrix and verify inversion ///////////////
    if (ditheFileName) {
        unsigned m2 = std::min(m/2, 8u);                                        // regular (square) stratification resolution
        fprintf(stderr, "m = %d, m2 = %d\n", m, m2);
        const int n = 1 << m2;
        const int N = n * n;
        uint32_t seqNo, mask((1 << m) - 1);
        std::vector<uint32_t> bitmap(N, N);                                     // N is invalid, so we use it to indicate uninitialized pixels.
        for (uint32_t seqNo = 0; seqNo < N; seqNo++) {
            Point p = xi[seqNo] >> (32 - m2);
            uint32_t pixelIndex = p.y * n + p.x;
            if (bitmap[pixelIndex] != N) {                                      // Already initialized?
                fprintf(
                    stderr,
                    "Two points, %d and %d, collided in pixel (%d, %d)\n",
                    seqNo, bitmap[pixelIndex], p.x, p.y
                );
            }
            bitmap[pixelIndex] = seqNo;
        }
        for (int Y = 0; Y < n; Y++) {
            for (int X = 0; X < n; X++) {
                seqNo = xi.getSeqNo(X, Y, m2);
                if (seqNo != bitmap[Y * n + X]) {
                    fprintf(stderr, "There is an error in inversion!\n");
                    exit(1);
                }
            }
        }
        FILE *file = fopen(ditheFileName, "w");
        fprintf(file, "P2\n%d %d\n%d\n", n, n, N - 1);
        for (int Y = 0; Y < n; Y++) {
            for (int X = 0; X < n; X++) {
                seqNo = xi.getSeqNo(X, Y, m2);
                fprintf(file, "%d\n", seqNo);
            }
        }
        fclose(file);
    }
    //////////////////////////////////////////////////////////////////////////*/

    //*/////////////////////// Report Generation Rate //////////////////////////
    if (reportGenerationRate) {
        uint32_t N = 1 << m;
        std::vector<Point> p(N);
        clock_t t0 = clock();
        for (uint32_t i = 0; i < N; i++) {
            p[i] = xi[i];
        }
        clock_t t1 = clock();
        double totalTime = (double)(t1 - t0) / CLOCKS_PER_SEC;
        fprintf(
            stderr, "done! Generated %10u points (2^%2u) "
            "in %9.6fs (%4dM points per second)\n",
                N, m, totalTime, (int)(N/(totalTime * 1000000))
        );
    }
    //////////////////////////////////////////////////////////////////////////*/

    //*/////////////////////// Report Inversion rate ///////////////////////////
    if (reportInversionRate ){                                                  // Localize variable names
        unsigned m2 = m/2;                                                      // regular (square) stratification resolution
        const uint32_t n = (1 << m2), N(n * n);
        std::vector<uint32_t> seqNo(N);
        clock_t t0 = clock();
        for (uint32_t Y = 0, i = 0; Y < n; Y++) {
            for (uint32_t X = 0; X < n; X++, i++) {
                seqNo[i] = xi.getSeqNo(X << (32 - m2), Y << (32 - m2));
                //seqNo[i] = xi.getSeqNo(X, Y, m2);
            }
        }
        clock_t t1 = clock();
        double totalTime = (double)(t1 - t0) / CLOCKS_PER_SEC;
        fprintf(
            stderr, "done! Inverted %10u strata (2^%2u) "
            "in %9.6fs (%4dM inversion per second)\n",
                N, 2 * m2, totalTime, (int)(N/(totalTime * 1000000))
        );
    }
    //////////////////////////////////////////////////////////////////////////*/


    //*///////////////////// Generate and print samples ////////////////////////
    if (outputFileName) {
        const Float inv = 1.0 / (1ULL << 32);
        uint32_t N = 1 << m;
        FILE *file = (
            outputFileName[0] == '-' ? stdout : fopen(outputFileName, "w")
        );
        fprintf(file, "%u\n", N);
        for (uint32_t i = 0; i < N; i++) {
            fprintf(file, "%0.20f %0.20f\n", inv * xi[i].x, inv * xi[i].y);
        }
        fclose(file);
    }
    //////////////////////////////////////////////////////////////////////////*/
}

